/* Threads.c -- multithreading library
2009-09-20 : Igor Pavlov : Public domain */

#ifndef _WIN32_WCE
#include <process.h>
#endif

#define SEM_IMPL
#include "Threads.h"

#ifdef __OS2__
#include <stdlib.h>
#include <stdio.h>

#if     DEBUG > 1
#define SEM_TRACE fprintf(stderr, "%s [%d] %s, handle=%08X->%08X->%08X\n",              \
                          __FILE__, __LINE__, __FUNCTION__, phandle, *(PULONG)phandle,  \
                          *(PULONG)phandle ? (*(SEM_SEMAPHORE**)phandle)->hevent : 0 ); \
                          fflush(stderr)
#else
#define SEM_TRACE
#endif

void OS2Abort(const char* message)
{
  PPIB ppib;
  DosGetInfoBlocks(NULL, &ppib);

  if (ppib->pib_ultype == 3) {
    WinMessageBox(HWND_DESKTOP, HWND_DESKTOP, message, "Critical Error", 100,
                  MB_OK | MB_ERROR | MB_APPLMODAL | MB_MOVEABLE);
  } else {
    fprintf(stderr, "%s\n", message);
    fflush(stderr);
  }
  abort();
}

typedef struct _THREADDATA {
  PFNTHREAD function;
  ULONG     param;
} THREADDATA;

static THREAD_FUNC_RET_TYPE THREAD_FUNC_CALL_TYPE Thread_Start(THREADDATA *pdata)
{
  HAB hab = WinInitialize(0);
  HMQ hmq = WinCreateMsgQueue(hab, 0);

  pdata->function(pdata->param);
  free(pdata);

  WinDestroyMsgQueue(hmq);
  WinTerminate(hab);
  return 0;
}

WRes Thread_Create(CThread *p, THREAD_FUNC_TYPE func, void* param)
{
  THREADDATA *pdata = malloc(sizeof(THREADDATA));

  if (pdata) {
    pdata->function = (PFNTHREAD)func;
    pdata->param = (ULONG)param;
    return DosCreateThread(p, (PFNTHREAD)Thread_Start, (ULONG)pdata, CREATE_READY, 1048576);
  } else {
    return ERROR_NOT_ENOUGH_MEMORY;
  }
}

BOOL Event_IsCreated(CEvent *phandle)
{
  SEM_EVENT *p = (SEM_EVENT*)*phandle;
  return p && p->hevent;
}

WRes Event_Close(CEvent *phandle)
{
  SEM_EVENT *p = (SEM_EVENT*)*phandle;
  SEM_TRACE;

  if (p) {
    *phandle = (CEvent)NULL;
    if (p->hevent) {
      DosCloseEventSem(p->hevent);
    }
    free(p);
  }

  return 0;
}

WRes Event_Wait(CEvent *phandle)
{
  SEM_EVENT *p = (SEM_EVENT*)*phandle;
  SEM_TRACE;

  if (p) {
    return DosWaitEventSem(p->hevent, SEM_INDEFINITE_WAIT);
  } else {
    return ERROR_INVALID_HANDLE;
  }
}

WRes Event_Set(CEvent *phandle)
{
  SEM_EVENT *p = (SEM_EVENT*)*phandle;
  WRes rc = ERROR_INVALID_HANDLE;

  SEM_TRACE;

  if (p) {
    rc = DosPostEventSem(p->hevent);
  }

  switch (rc) {
    case ERROR_ALREADY_POSTED:
    case ERROR_TOO_MANY_POSTS:
      rc = NO_ERROR;
  }
  return rc;
}

WRes Event_Reset(CEvent *phandle)
{
  SEM_EVENT *p = (SEM_EVENT*)*phandle;
  WRes rc = ERROR_INVALID_HANDLE;
  ULONG ulPostCt;

  SEM_TRACE;

  if (p) {
    rc = DosResetEventSem(p->hevent, &ulPostCt);
  }
  switch (rc) {
    case ERROR_ALREADY_RESET:
      rc = NO_ERROR;
  }
  return rc;
}

WRes ManualResetEvent_Create(CManualResetEvent *phandle, int signaled)
{
  SEM_EVENT *p = calloc(sizeof(SEM_EVENT), 1);
  WRes rc;

  if (p) {
    if ((rc = DosCreateEventSem(NULL, &p->hevent, 0, signaled)) == 0) {
      p->type = OS2_EVENT;
     *phandle = (CEvent)p;
    } else {
      free(p);
    }
  } else {
    rc = ERROR_NOT_ENOUGH_MEMORY;
  }

  SEM_TRACE;
  return rc;
}

WRes ManualResetEvent_CreateNotSignaled(CManualResetEvent *phandle) {
  return ManualResetEvent_Create(phandle, 0);
}

WRes AutoResetEvent_Create(CAutoResetEvent *phandle, int signaled)
{
  SEM_EVENT *p = calloc(sizeof(SEM_EVENT), 1);
  WRes rc;

  if (p) {
    if ((rc = DosCreateEventSem(NULL, &p->hevent, DCE_POSTONE, signaled)) == 0) {
      p->type = OS2_EVENT;
     *phandle = (CEvent)p;
    } else {
      free(p);
    }
  } else {
    rc = ERROR_NOT_ENOUGH_MEMORY;
  }

  SEM_TRACE;
  return rc;
}

WRes AutoResetEvent_CreateNotSignaled(CAutoResetEvent *p) {
  return AutoResetEvent_Create(p, 0);
}

WRes CriticalSection_Init(CCriticalSection *phandle)
{
  SEM_MUTEX *p = calloc(sizeof(SEM_MUTEX), 1);
  WRes rc;

  if (p) {
    if ((rc = DosCreateMutexSem(NULL, &p->hmutex, 0, FALSE)) == 0) {
      p->type = OS2_MUTEX;
     *phandle = (CCriticalSection)p;
    }
  } else {
    rc = ERROR_NOT_ENOUGH_MEMORY;
  }

  SEM_TRACE;
  return rc;
}

void CriticalSection_Delete(CCriticalSection *phandle)
{
  SEM_MUTEX *p = (SEM_MUTEX*)*phandle;
  SEM_TRACE;

  if (p) {
    *phandle = (CCriticalSection)NULL;
    if (p->hmutex) {
      DosCloseMutexSem(p->hmutex);
    }
    free(p);
  }
}

WRes CriticalSection_Enter(CCriticalSection *phandle)
{
  SEM_MUTEX *p = (SEM_MUTEX*)*phandle;
  SEM_TRACE;

  if (p) {
    return DosRequestMutexSem(p->hmutex, SEM_INDEFINITE_WAIT);
  } else {
    return ERROR_INVALID_HANDLE;
  }
}

WRes CriticalSection_Leave(CCriticalSection *phandle)
{
  SEM_MUTEX *p = (SEM_MUTEX*)*phandle;
  SEM_TRACE;

  if (p) {
    return DosReleaseMutexSem(p->hmutex);
  } else {
    return ERROR_INVALID_HANDLE;
  }
}

WRes Semaphore_Create(CSemaphore *phandle, UInt32 initCount, UInt32 maxCount)
{
  SEM_SEMAPHORE *p;
  WRes rc;

  if (initCount > maxCount || maxCount == 0) {
    return ERROR_INVALID_PARAMETER;
  }

  p = calloc(sizeof(SEM_SEMAPHORE), 1);

  if (p) {
    p->type = OS2_SEMAPHORE;
    p->count = initCount;
    p->countmax = maxCount;

    if((rc = DosCreateMutexSem(NULL, &p->hmutex, 0, 0)) != 0) {
      free(p);
      return rc;
    }
    if((rc = DosCreateEventSem(NULL, &p->hevent, DCE_POSTONE, initCount > 0)) != 0) {
      DosCloseMutexSem(p->hmutex);
      free(p);
      return rc;
    }

   *phandle = (CSemaphore)p;
  } else {
    rc = ERROR_NOT_ENOUGH_MEMORY;;
  }

  SEM_TRACE;
  return rc;
}

WRes Semaphore_Close(CSemaphore *phandle)
{
  SEM_SEMAPHORE *p = (SEM_SEMAPHORE*)*phandle;
  SEM_TRACE;

  if (p) {
    *phandle = (CSemaphore)NULL;
    if (p->hevent) {
      DosCloseEventSem(p->hevent);
    }
    if (p->hmutex) {
      DosCloseMutexSem(p->hmutex);
    }
    free(p);
  }

  return 0;
}

static WRes Semaphore_Leave(SEM_SEMAPHORE *p)
{
  WRes rc;

  if (p) {
    if ((rc = DosRequestMutexSem(p->hmutex, SEM_INDEFINITE_WAIT)) != 0) {
      return rc;
    }
    if (--p->count) {
      DosPostEventSem(p->hevent);
    }
    rc = DosReleaseMutexSem(p->hmutex);
  } else {
    rc = ERROR_INVALID_HANDLE;
  }

  return rc;
}

WRes Semaphore_Wait(CSemaphore *phandle)
{
  SEM_SEMAPHORE *p = (SEM_SEMAPHORE*)*phandle;
  WRes rc;

  SEM_TRACE;

  if (p) {
    if ((rc = DosWaitEventSem(p->hevent, SEM_INDEFINITE_WAIT)) != 0) {
      return rc;
    }
    rc = Semaphore_Leave(p);
  } else {
    rc = ERROR_INVALID_HANDLE;
  }

  return rc;
}

WRes Semaphore_ReleaseN(CSemaphore *phandle, UInt32 num)
{
  SEM_SEMAPHORE *p = (SEM_SEMAPHORE*)*phandle;
  WRes rc;

  SEM_TRACE;

  if (p) {
    if ((rc = DosRequestMutexSem(p->hmutex, SEM_INDEFINITE_WAIT)) != 0) {
      return rc;
    }
    if (p->count + num <= p->countmax) {
      p->count += num;
      DosPostEventSem(p->hevent);
    } else {
      rc = ERROR_TOO_MANY_POSTS;
    }
    DosReleaseMutexSem(p->hmutex);
  } else {
    rc = ERROR_INVALID_HANDLE;
  }

  return rc;
}

WRes Semaphore_Release1(CSemaphore *p) {
  return Semaphore_ReleaseN(p, 1);
}

DWORD WaitForMultipleObjects(DWORD count, const HANDLE *handles, BOOL wait_all, DWORD timeout)
{
  DWORD i;
  WRes rc;

  #ifdef DEBUG
  fprintf(stderr, "%s [%d] %s, handles=", __FILE__, __LINE__, __FUNCTION__);
  for (i = 0; i < count; i++) {
    fprintf(stderr, "%08X ", handles[i]);
    if (handles[i]) {
      switch(((SEM_SEMAPHORE*)handles[i])->type) {
        case OS2_EVENT:
          fprintf(stderr, "[EVENT:%08X] ", ((SEM_SEMAPHORE*)handles[i])->hevent);
          break;
        case OS2_MUTEX:
          fprintf(stderr, "[MUTEX:%08X] ", ((SEM_SEMAPHORE*)handles[i])->hevent);
          break;
        case OS2_SEMAPHORE:
          fprintf(stderr, "[SEM:%08X] ", ((SEM_SEMAPHORE*)handles[i])->hevent);
          break;
        default:
          fprintf(stderr, "[UNKNOWN] ");
          break;
      }
    } else {
      fprintf(stderr, "[NULL] ");
    }
  }
  fprintf(stderr, "\n");
  fflush(stderr);
  #endif

  for(;;) {
    for (i = 0; i < count; i++) {
      switch(((SEM_SEMAPHORE*)handles[i])->type) {
        case OS2_EVENT:
          if ((rc = DosWaitEventSem(((SEM_EVENT*)handles[i])->hevent, SEM_IMMEDIATE_RETURN)) == 0) {
            return WAIT_OBJECT_0 + i;
          }
          break;
        case OS2_MUTEX:
          if ((rc = DosRequestMutexSem(((SEM_MUTEX*)handles[i])->hmutex, SEM_IMMEDIATE_RETURN)) == 0) {
            return WAIT_OBJECT_0 + i;
          }
          break;
        case OS2_SEMAPHORE:
          if ((rc = DosWaitEventSem(((SEM_SEMAPHORE*)handles[i])->hevent, SEM_IMMEDIATE_RETURN)) == 0) {
            Semaphore_Leave((SEM_SEMAPHORE*)handles[i]);
            return WAIT_OBJECT_0 + i;
          }
          break;
      }
      if (rc != ERROR_TIMEOUT) {
        errno = rc;
        return 0xFFFFFFFF;
      }
      DosSleep(0);
    }
  }
}

#else /* __OS2__ */

static WRes GetError()
{
  DWORD res = GetLastError();
  return (res) ? (WRes)(res) : 1;
}

WRes HandleToWRes(HANDLE h) { return (h != 0) ? 0 : GetError(); }
WRes BOOLToWRes(BOOL v) { return v ? 0 : GetError(); }

WRes HandlePtr_Close(HANDLE *p)
{
  if (*p != NULL)
    if (!CloseHandle(*p))
      return GetError();
  *p = NULL;
  return 0;
}

WRes Handle_WaitObject(HANDLE h) { return (WRes)WaitForSingleObject(h, INFINITE); }

WRes Thread_Create(CThread *p, THREAD_FUNC_TYPE func, LPVOID param)
{
  unsigned threadId; /* Windows Me/98/95: threadId parameter may not be NULL in _beginthreadex/CreateThread functions */
  *p =
    #ifdef UNDER_CE
    CreateThread(0, 0, func, param, 0, &threadId);
    #else
    (HANDLE)_beginthreadex(NULL, 0, func, param, 0, &threadId);
    #endif
    /* maybe we must use errno here, but probably GetLastError() is also OK. */
  return HandleToWRes(*p);
}

WRes Event_Create(CEvent *p, BOOL manualReset, int signaled)
{
  *p = CreateEvent(NULL, manualReset, (signaled ? TRUE : FALSE), NULL);
  return HandleToWRes(*p);
}

WRes Event_Set(CEvent *p) { return BOOLToWRes(SetEvent(*p)); }
WRes Event_Reset(CEvent *p) { return BOOLToWRes(ResetEvent(*p)); }

WRes ManualResetEvent_Create(CManualResetEvent *p, int signaled) { return Event_Create(p, TRUE, signaled); }
WRes AutoResetEvent_Create(CAutoResetEvent *p, int signaled) { return Event_Create(p, FALSE, signaled); }
WRes ManualResetEvent_CreateNotSignaled(CManualResetEvent *p) { return ManualResetEvent_Create(p, 0); }
WRes AutoResetEvent_CreateNotSignaled(CAutoResetEvent *p) { return AutoResetEvent_Create(p, 0); }


WRes Semaphore_Create(CSemaphore *p, UInt32 initCount, UInt32 maxCount)
{
  *p = CreateSemaphore(NULL, (LONG)initCount, (LONG)maxCount, NULL);
  return HandleToWRes(*p);
}

static WRes Semaphore_Release(CSemaphore *p, LONG releaseCount, LONG *previousCount)
  { return BOOLToWRes(ReleaseSemaphore(*p, releaseCount, previousCount)); }
WRes Semaphore_ReleaseN(CSemaphore *p, UInt32 num)
  { return Semaphore_Release(p, (LONG)num, NULL); }
WRes Semaphore_Release1(CSemaphore *p) { return Semaphore_ReleaseN(p, 1); }

WRes CriticalSection_Init(CCriticalSection *p)
{
  /* InitializeCriticalSection can raise only STATUS_NO_MEMORY exception */
  #ifdef _MSC_VER
  __try
  #endif
  {
    InitializeCriticalSection(p);
    /* InitializeCriticalSectionAndSpinCount(p, 0); */
  }
  #ifdef _MSC_VER
  __except (EXCEPTION_EXECUTE_HANDLER) { return 1; }
  #endif
  return 0;
}

#endif /* __OS2__ */

